/*
 * ext_all.c - Dump entire configuration at external interface
 *
 * Written 2001 by Werner Almesberger 
 * Copyright 2001 Network Robots
 */

/*
 * This is experimental code and will change. Probably a lot.
 *
 * Known problems: @@@
 * - should allow simple configurations using other qdiscs than dsmark
 * - does not support qdiscs shared by multiples classes
 * - gets confused if dumping more than one qdisc tree (ingress/egress) per
 *   session
 * - does not take into account that only part of the dsmark class may be used
 *   for marking
 * - does not take into account that no dsmark class may exist for marking
 *   (in fact, should look at parameters to decide whether a change is
 *   actually requested)
 * - does not report inner qdiscs changing skb->tc_index
 */


#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "util.h"
#include "error.h"
#include "param.h"
#include "tree.h"
#include "filter.h"
#include "qdisc.h"
#include "device.h"
#include "if.h"
#include "iflib.h"
#include "ext.h"
#include "ext_all.h"


int dump_all; /* dump all qdiscs and filters (experimental) */


static const QDISC *_root; /* @@@@ HACK */


/* -------------------- skb->tc_index to class mapping --------------------- */


static void print_class_n(FILE *file,const QDISC *qdisc,uint16_t class,
  int *first)
{
    fprintf(file,"%c%u:%u",*first ? ' ' : ',',(unsigned) qdisc->number,
      (unsigned) class);
    *first = 0;
}


static void print_class(FILE *file,const CLASS *class,int *first)
{
    print_class_n(file,class->parent.qdisc,class->number,first);
}


static void gred_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    const CLASS *class;

    param_get(qdisc->params,qdisc->location);
    for (class = qdisc->classes; class; class = class->next)
	if (class->number == (index & 0xf)) {
	    print_class(file,class,&first);
	    return;
	}
    print_class_n(file,qdisc,prm_default_index.v,&first);
}


static CLASS *find_class(CLASS *classes,uint32_t number)
{
    CLASS *class;

    for (class = classes; class; class = class->next) {
	if (class->number == number) return class;
	if (class->child) {
	    CLASS *found;

	    found = find_class(class->child,number);
	    if (found) return found;
	}
    }
    return NULL;
}


static void find_path(FILE *file,const QDISC *qdisc,uint16_t index,int first);


static uint32_t filter_path(const FILTER *filters,uint16_t index,int first)
{
    const FILTER *filter;

    for (filter = filters; filter; filter = filter->next) {
	const ELEMENT *element;

	if (filter->dsc != &tcindex_dsc)
	    yyerrorf("dump_all supports only \"tcindex\", not \"%s\"",
	      filter->dsc->name);
	for (element = filter->elements; element; element = element->next)
	    if (element->police)
		yyerror("dump_all does not support policing at tcindex");
    }
    /*
     * Note: filter.c ensures that filters are sorted by priority, with the
     * hightest priority (lowest number) first.
     */
    for (filter = filters; filter; filter = filter->next) {
	const ELEMENT *element;
	int shift,mask;
	unsigned value;

	param_load(filter->params,filter->dsc->filter_param,
	  filter->dsc->filter_param,filter->location);
	shift = prm_shift.present ? prm_shift.v : 0;
	mask = prm_mask.present ? prm_mask.v : 0xffff;
	value = (index & mask) >> shift;
	debugf("(index 0x%x & mask 0x%x) >> shift %d -> 0x%x",(unsigned) index,
	  mask,shift,value);
	for (element = filter->elements; element; element = element->next)
	    if ((index & mask) >> shift == element->number)
		return element->parent.class->number;
	if (!prm_fall_through.present || !prm_fall_through.v) continue;
	return value;
    }
    return UNDEF_U32;
}


static void qdisc_filter_path(FILE *file,const QDISC *qdisc,uint16_t index,
  int first)
{
    const FILTER *filters = qdisc->filters;
    const CLASS *class;
    uint32_t class_num;

    if (!filters) {
	if (qdisc->dsc->flags & QDISC_HAS_CLASSES)
	    yyerrorf("don't know how to select classes of \"%s\"",
	      qdisc->dsc->name);
	return;
    }
    if (prm_present(qdisc->params,&prm_priomap))
	yyerrorf("dump_all does not support the \"priomap\" parameter");
    class_num = filter_path(filters,index,first);
    if (class_num == UNDEF_U32) {
	if (qdisc->dsc == &cbq_dsc) class_num = 0;
	else yyerrorf("dump_all found no matching choice for value 0x%x in "
	  "tcindex",(unsigned) index);
    }
    while (1) {
	uint32_t next_class;

	class = find_class(qdisc->classes,class_num);
	if (!class) break;
	if (prm_present(class->params,&prm_priomap))
	    yyerrorf("dump_all does not support the \"priomap\" parameter");
	if (!class->filters) break;
	next_class = filter_path(class->filters,index,first);
	if (next_class == UNDEF_U32) break;
	class_num = next_class;
    }
    print_class_n(file,qdisc,class_num,&first);
    if (class && class->qdisc) find_path(file,class->qdisc,index,first);
}


static void find_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    if (qdisc->dsc == &gred_dsc) gred_path(file,qdisc,index,first);
    else qdisc_filter_path(file,qdisc,index,first);
}


int dump_all_decision(FILE *file,uint16_t class)
{
    const CLASS *c;

    if (!dump_all) return 0;
    for (c = _root->classes; c; c = c->next)
	if (c->qdisc) {
	    int one = 1;

	    fprintf(file," class");
	    print_class_n(file,_root,class,&one);
	    find_path(file,c->qdisc,class,0);
	    return 1;
	}
    return 0;
}


/* ------------------- Dumps qdisc and class definitions ------------------- */


static void dump_this_qdisc(FILE *file,QDISC *qdisc);


static void dump_qdiscs_at_classes(FILE *file,const CLASS *classes)
{
    const CLASS *class;

    for (class = classes; class; class = class->next) {
	if (class->qdisc) dump_this_qdisc(file,class->qdisc);
	dump_qdiscs_at_classes(file,class->child);
    }
}


static void dump_this_qdisc(FILE *file,QDISC *qdisc)
{
    dump_qdiscs_at_classes(file,qdisc->classes);
    if (!qdisc->dsc->dump_ext)
	yyerrorf("don't know how to dump_ext qdisc \"%s\"",qdisc->dsc->name);
    qdisc->dsc->dump_ext(file,qdisc);
}


/* ------------------------------------------------------------------------- */


static void prepare_block(QDISC *root)
{
    if (!root->filters) return;
    root->if_expr = iflib_combine(root);
    dump_if_ext_prepare(root->filters);
}


static void generate_block(FILE *file,QDISC *root)
{
    dump_block(file,root);
    dump_this_qdisc(file,root);
    if (root->filters) dump_if_ext_local(root->filters,file);
}


static void dump_all_callback(FILE *file,void *user)
{
    const DEVICE *device;

    dump_pragma(file);
    for (device = devices; device; device = device->next) {
	if (device->ingress) prepare_block(device->ingress);
	if (device->egress) prepare_block(device->egress);
    }
    dump_if_ext_global(file);
    for (device = devices; device; device = device->next) {
	if (device->ingress) generate_block(file,device->ingress);
	if (device->egress) generate_block(file,device->egress);
    }
    reset_offsets();
}


static void no_root_class_filters(const CLASS *classes)
{
    const CLASS *class;

    for (class = classes; class; class = class->next) {
	if (class->filters)
	    yyerrorf("dump_all: classes of root qdisc may not have filters");
	no_root_class_filters(class->child);
    }
}


static void no_dsmark_constraints(const CLASS *classes)
{
    const CLASS *class;

    for (class = classes; class; class = class->next) {
	if (class->qdisc &&
	  (class->qdisc->dsc->flags & (QDISC_HAS_CLASSES | QDISC_HAS_FILTERS)))
	    yyerror("dump_all: if not dsmark, only root qdisc may have "
	      "classes");
	no_dsmark_constraints(class->child);
    }
}


static void check_one_qdisc(const QDISC *root)
{
    if (!root) yyerror("dump_all: no root qdisc");
    _root = root;
    if (root->filters) {
	if (root->filters->next) yyerror("dump_all: only one filter allowed");
	if (root->filters->dsc != &if_dsc)
	    yyerrorf("dump_all: filter must be \"if\", not \"%s\"",
	      root->filters->dsc->name);
    }
    no_root_class_filters(root->classes);
    if (root->dsc != &dsmark_dsc) no_dsmark_constraints(root->classes);
}


void dump_all_devices(const char *target)
{
    const DEVICE *device;

    for (device = devices; device; device = device->next) {
	if (device->ingress) check_one_qdisc(device->ingress);
	if (device->egress) check_one_qdisc(device->egress);
    }
    ext_build(target,NULL,NULL,dump_all_callback,NULL);
}


void dump_one_qdisc(QDISC *root,const char *target)
{
    error("not (yet ?) implemented");
}
